@@ -31,6 +31,7 @@
 #include <orbit/orbit.h>
 
 #include "cddb-slave.h"
+#include "cddb-parser.h"
 #include "GNOME_Media_CDDBSlave2.h"
 
 
@@ -98,8 +99,12 @@ typedef struct _ConnectionData {
 
 static GHashTable *pending_requests = NULL;
 static GHashTable *cddb_cache = NULL;
+static GList *cddb_slaves = NULL;
 
 static void do_hello (ConnectionData *cd);
+
+extern gboolean cddb_debugging;
+
 
 #define CDDB_SLAVE_CDDB_FINISHED "GNOME_Media_CDDBSlave2:CDDB-Finished"
 
@@ -112,7 +117,8 @@ cddb_slave_notify_listeners (CDDBSlave *cddb,
 {
 	CORBA_any any;
 	GNOME_Media_CDDBSlave2_QueryResult qr;
-
+	GList *l;
+	
 	g_return_if_fail (cddb != NULL);
 	g_return_if_fail (IS_CDDB_SLAVE (cddb));
 
@@ -122,9 +128,19 @@ cddb_slave_notify_listeners (CDDBSlave *cddb,
 	any._type = (CORBA_TypeCode) TC_GNOME_Media_CDDBSlave2_QueryResult;
 	any._value = &qr;
 
-	bonobo_event_source_notify_listeners (cddb->priv->event_source,
-					      CDDB_SLAVE_CDDB_FINISHED,
-					      &any, NULL);
+	if (cddb_debugging == TRUE) {
+		g_print ("CDDB: Notifying listeners.\n");
+	}
+
+	for (l = cddb_slaves; l; l = l->next) {
+		CDDBSlave *slave;
+
+		slave = CDDB_SLAVE (l->data);
+		
+		bonobo_event_source_notify_listeners (slave->priv->event_source,
+						      CDDB_SLAVE_CDDB_FINISHED,
+						      &any, NULL);
+	}
 }
 
 static gboolean
@@ -137,13 +153,17 @@ do_goodbye_response (ConnectionData *cd,
 	code = atoi (response);
 	switch (code) {
 	case 230:
-		g_print ("Disconnected\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Disconnected\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		break;
 
 	default:
-		g_print ("Unknown response\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Unknown response\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		break;
 	}
 
@@ -188,7 +208,6 @@ clear_entry_from_cache (const char *discid)
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	/* Set the fields of the entry to unknown */
 	entry->revision = 0;
-	entry->topline = NULL;
 
 	g_hash_table_insert (entry->fields, g_strdup ("DTITLE"),
 			     g_string_new (_("Unknown / Unknown")));
@@ -212,6 +231,11 @@ do_goodbye (ConnectionData *cd)
 	/* Send quit command */
 	status = gnet_io_channel_writen (cd->iochannel, "quit\n",
 					 5, &bytes_writen);
+	if (cddb_debugging == TRUE) {
+		g_print ("CDDB: status: %d bytes_writen %d\n",
+			 status, bytes_writen);
+	}
+
 	cd->mode = CONNECTION_MODE_NEED_GOODBYE;
 }
 
@@ -234,6 +258,11 @@ do_read_response (ConnectionData *cd,
 
 	switch (code) {
 	case 210:
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Received data.\n");
+			g_print ("CDDB: %s\n", response);
+		}
+		
 		if (waiting_for_terminator == FALSE) {
 			/* Open the file */
 			char *filename, *dirname;
@@ -242,22 +271,31 @@ do_read_response (ConnectionData *cd,
 			filename = g_concat_dir_and_file (dirname, cd->discid);
 			g_free (dirname);
 
-			g_print ("Opening %s\n", filename);
+			if (cddb_debugging == TRUE) {
+				g_print ("CDDB: Opening %s\n", filename);
+			}
+			
 			handle = fopen (filename, "w");
-			g_free (filename);
 			
 			if (handle == NULL) {
-				g_print ("Erk!\n");
+				g_warning ("Could not open %s\n", filename);
+				g_free (filename);
 				return FALSE;
 			}
+			g_free (filename);
 
 			waiting_for_terminator = TRUE;
-		}
+		} else {
 
-		g_assert (handle != NULL);
+			g_assert (handle != NULL);
 		
-		/* Write the line */
-		fputs (response, handle);
+			/* Write the line */
+			if (cddb_debugging == TRUE) {
+				g_print ("CDDB: Writing %s\n", response);
+			}
+			
+			fputs (response, handle);
+		}
 
 		if (response[0] == '.') {
 			CDDBEntry *entry;
@@ -271,6 +309,9 @@ do_read_response (ConnectionData *cd,
 			g_free (dirname);
 			
 			entry = g_hash_table_lookup (cddb_cache, cd->discid);
+			if (cddb_debugging == TRUE) {
+				g_print ("CDDB: Parsing %s\n", filename);
+			}
 			cddb_entry_parse_file (entry, filename);
 			g_free (filename);
 			
@@ -286,32 +327,40 @@ do_read_response (ConnectionData *cd,
 		break;
 
 	case 401:
-		g_print ("Specified CDDB entry not found\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Specified CDDB entry not found\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		more = FALSE;
 		disconnect = TRUE;
 		clear_entry_from_cache (cd->discid);
 		break;
 
 	case 402:
-		g_print ("Server error\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Server error\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		more = FALSE;
 		disconnect = TRUE;
 		clear_entry_from_cache (cd->discid);
 		break;
 
 	case 403:
-		g_print ("Database entry is corrupt\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Database entry is corrupt\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		more = FALSE;
 		disconnect = TRUE;
 		clear_entry_from_cache (cd->discid);
 		break;
 
 	case 409:
-		g_print ("No handshake\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: No handshake\n");
+			g_print ("CDDB: %s\n", response);
+		}
 
 		/* Handshake */
 		do_hello (cd);
@@ -341,6 +390,10 @@ do_read (ConnectionData *cd,
 	query = g_strdup_printf ("cddb read %s %s\n", cat, discid);
 	status = gnet_io_channel_writen (cd->iochannel, query,
 					 strlen (query), &bytes_writen);
+	if (cddb_debugging == TRUE) {
+		g_print ("CDDB: status: %d bytes_writen %d\n",
+			 status, bytes_writen);
+	}
 	g_free (query);
 	cd->mode = CONNECTION_MODE_NEED_READ_RESPONSE;
 }
@@ -356,9 +409,16 @@ create_model_from_list (GList *list)
 
 	for (l = list; l; l = l->next) {
 		char **vector;
-
+		char *end;
+		
 		vector = (char **) l->data;
-
+		/* Strip newlines and \r */
+		vector[2][strlen (vector[2]) - 1] = 0;
+		end = strchr (vector[2], '\r');
+		if (end != NULL) {
+			*end = 0;
+		}
+		
 		gtk_list_store_append (store, &iter);
 		gtk_list_store_set (store, &iter, 0, vector[0], 1, vector[1], 2, vector[2], -1);
 	}
@@ -385,7 +445,7 @@ display_results (ConnectionData *cd)
 	label = gtk_label_new (_("There were multiple matches found in the database.\n"
 				 "Below is a list of possible matches, please choose the "
 				 "best match"));
-	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->vbox), label, TRUE, TRUE, 0);
+	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->vbox), label, FALSE, FALSE, 0);
 	gtk_widget_show (label);
 	
 	model = create_model_from_list (cd->matches);
@@ -464,12 +524,16 @@ do_query_response (ConnectionData *cd,
 	
 	switch (code) {
 	case 200:
-		g_print ("Exact match found.\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Exact match found.\n");
+			g_print ("CDDB: %s\n", response);
+		}
 
 		vector = g_strsplit (response, " ", 4);
 		if (vector == NULL) {
-			g_print ("Erk!\n");
+			if (cddb_debugging == TRUE) {
+				g_print ("CDDB: Response did not contain all the information.\n");
+			}
 			return FALSE;
 		}
 
@@ -483,7 +547,12 @@ do_query_response (ConnectionData *cd,
 		break;
 
 	case 211:
-
+		/* Should this be 210 as well? */
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Multiple matches found.\n");
+			g_print ("CDDB: %s\n", response);
+		}
+		
 		if (response[0] == '.') {
 			/* Terminator */
 			char **result;
@@ -529,7 +598,9 @@ do_query_response (ConnectionData *cd,
 		if (waiting_for_terminator == TRUE) {
 			vector = g_strsplit (response, " ", 3);
 			if (vector == NULL) {
-				g_print ("Erk!\n");
+				if (cddb_debugging == TRUE) {
+					g_print ("CDDB: Response did not containe enough info\n");
+				}
 				return FALSE;
 			}
 			
@@ -543,30 +614,38 @@ do_query_response (ConnectionData *cd,
 		break;
 
 	case 202:
-		g_print ("No match found\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: No match found\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		more = FALSE;
 		disconnect = TRUE;
 		break;
 
 	case 403:
-		g_print ("Database entry is corrupt\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Database entry is corrupt\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		more = FALSE;
 		disconnect = TRUE;
 		break;
 
 	case 409:
-		g_print ("No handshake\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: No handshake\n");
+			g_print ("CDDB: %s\n", response);
+		}
 
 		do_hello (cd);
 		more = FALSE;
 		break;
 
 	default:
-		g_print ("Unknown response\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Unknown response\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		disconnect = TRUE;
 		more = FALSE;
 		break;
@@ -603,7 +682,10 @@ do_query (ConnectionData *cd)
 				 cd->offsets, cd->nsecs);
 	status = gnet_io_channel_writen (cd->iochannel, query,
 					 strlen (query), &bytes_writen);
-	g_print ("status: %d bytes_writen %d\n", status, bytes_writen);
+	if (cddb_debugging == TRUE) {
+		g_print ("CDDB: status: %d bytes_writen %d\n",
+			 status, bytes_writen);
+	}
 	g_free (query);
 	cd->mode = CONNECTION_MODE_NEED_QUERY_RESPONSE;
 }
@@ -618,13 +700,17 @@ do_hello_response (ConnectionData *cd,
 	code = atoi (response);
 	switch (code) {
 	case 200:
-		g_print ("Hello ok - Welcome\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Hello ok - Welcome\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		break;
 
 	case 431:
-		g_print ("Hello unsuccessful\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Hello unsuccessful\n");
+			g_print ("CDDB: %s\n", response);
+		}
 
 		/* Disconnect here */
 		clear_entry_from_cache (cd->discid);
@@ -632,13 +718,17 @@ do_hello_response (ConnectionData *cd,
 		break;
 
 	case 402:
-		g_print ("Already shook hands\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Already shook hands\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		break;
 
 	default:
-		g_print ("Unknown response\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Unknown response\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		break;
 	}
 
@@ -668,7 +758,10 @@ do_hello (ConnectionData *cd)
 	/* Need to check the return of this one */
 	status = gnet_io_channel_writen (cd->iochannel, hello,
 					 strlen (hello), &bytes_writen);
-	g_print ("Status: %d bytes_writen: %d\n", status, bytes_writen);
+	if (cddb_debugging == TRUE) {
+		g_print ("CDDB: Status: %d bytes_writen: %d\n",
+			 status, bytes_writen);
+	}
 	g_free (hello);
 
 	cd->mode = CONNECTION_MODE_NEED_HELLO_RESPONSE;
@@ -685,38 +778,51 @@ do_open_response (ConnectionData *cd,
 	code = atoi (response);
 	switch (code) {
 	case 200:
-		g_print ("Hello ok - Read/Write access allowed\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Hello ok - Read/Write access allowed\n");
+			g_print ("CDDB: %s\n", response);
+		}
+		
 		cddb->priv->access = CDDB_ACCESS_READWRITE;
 		break;
 		
 	case 201:
-		g_print ("Hello ok - Read only access\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Hello ok - Read only access\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		cddb->priv->access = CDDB_ACCESS_READONLY;
 		break;
 		
 	case 432:
-		g_print ("No more connections allowed\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: No more connections allowed\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		cddb->priv->access = CDDB_ACCESS_NONE;
 		break;
 
 	case 433:
-		g_print ("No connections allowed: X users allowed, Y currently active\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: No connections allowed: X users allowed, Y currently active\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		cddb->priv->access = CDDB_ACCESS_NONE;
 		break;
 
 	case 434:
-		g_print ("No connections allowed: system load too high\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: No connections allowed: system load too high\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		cddb->priv->access = CDDB_ACCESS_NONE;
 		break;
 
 	default:
-		g_print ("Unknown response code: %d\n");
-		g_print ("%s\n", response);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Unknown response code: %d\n");
+			g_print ("CDDB: %s\n", response);
+		}
 		cddb->priv->access = CDDB_ACCESS_NONE;
 		break;
 	}
@@ -837,6 +943,10 @@ open_cb (GTcpSocket *sock,
 
 	sin = gnet_tcp_socket_get_iochannel (sock);
 	cd->iochannel = sin;
+
+	if (cddb_debugging == TRUE) {
+		g_print ("CDDB: Opened connection, adding watch\n");
+	}
 	
 	cd->tag = g_io_add_watch (sin, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 				  read_from_server, data);
@@ -847,6 +957,12 @@ cddb_send_cmd (ConnectionData *data)
 {
 	GTcpSocketConnectAsyncID *sock;
 
+	if (cddb_debugging == TRUE) {
+		g_print ("CDDB: Opening socket to %s on port %d",
+			 data->cddb->priv->server,
+			 data->cddb->priv->port);
+	}
+	
 	sock = gnet_tcp_socket_connect_async (data->cddb->priv->server,
 					      data->cddb->priv->port,
 					      open_cb, data);
@@ -906,10 +1022,17 @@ impl_GNOME_Media_CDDBSlave2_query (PortableServer_Servant servant,
 	char *request, *safe_offsets;
 	char *username, *hostname, *fullname, *uri;
 
+	if (cddb_debugging == TRUE) {
+		g_print ("CDDB: Querying for %s\n", discid);
+	}
 	cddb = cddb_slave_from_servant (servant);
 
 	if (cddb_check_cache (discid) == TRUE) {
-		cddb_slave_notify_listeners (cddb, discid, GNOME_Media_CDDBSlave2_OK);
+		if (cddb_debugging == TRUE) {
+			g_print ("CDDB: Found %s in cache\n", discid);
+		}
+		cddb_slave_notify_listeners (cddb, discid,
+					     GNOME_Media_CDDBSlave2_OK);
 		return;
 	}
 
@@ -935,6 +1058,35 @@ impl_GNOME_Media_CDDBSlave2_query (PortableServer_Servant servant,
 	cddb_send_cmd (cd);
 }
 
+static void
+impl_GNOME_Media_CDDBSlave2_save (PortableServer_Servant servant,
+				  const CORBA_char *discid,
+				  CORBA_Environment *ev)
+{
+	CDDBSlave *cddb;
+	CDDBEntry *entry;
+	gboolean write_ret;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	write_ret = cddb_entry_write_to_file (entry);
+	if (write_ret == FALSE) {
+		/* Exception */
+		return;
+	}
+
+	cddb = cddb_slave_from_servant (servant);
+	/* FIXME: Should we only emit the changed signal here,
+	   or in the set* calls? */
+	cddb_slave_notify_listeners (cddb, discid, GNOME_Media_CDDBSlave2_OK);
+}
+
 static CORBA_char *
 impl_GNOME_Media_CDDBSlave2_getArtist (PortableServer_Servant servant,
 				       const CORBA_char *discid,
@@ -947,13 +1099,17 @@ impl_GNOME_Media_CDDBSlave2_getArtist (PortableServer_Servant servant,
 	
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
-		/* Should set an exception here */
-		return NULL;
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return CORBA_string_dup ("Unknown artist");
 	}
 
 	dtitle = g_hash_table_lookup (entry->fields, "DTITLE");
 	if (dtitle == NULL) {
-		return NULL;
+		g_warning ("No field DTITLE %s", discid);
+		return CORBA_string_dup ("Unknown artist");
 	}
 
 	split = strstr (dtitle->str, " / ");
@@ -980,6 +1136,9 @@ impl_GNOME_Media_CDDBSlave2_setArtist (PortableServer_Servant servant,
 
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
 		return;
 	}
 
@@ -1009,12 +1168,17 @@ impl_GNOME_Media_CDDBSlave2_getDiscTitle (PortableServer_Servant servant,
 
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
-		return NULL;
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return CORBA_string_dup ("Unknown disc");
 	}
 
 	dtitle = g_hash_table_lookup (entry->fields, "DTITLE");
 	if (dtitle == NULL) {
-		return NULL;
+		g_warning ("Entry has no DTITLE %s", discid);
+		return CORBA_string_dup ("Unknown disc");
 	}
 
 	/* FIXME: use g_strsplit */
@@ -1038,6 +1202,9 @@ impl_GNOME_Media_CDDBSlave2_setDiscTitle (PortableServer_Servant servant,
 
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
 		return;
 	}
 
@@ -1064,6 +1231,10 @@ impl_GNOME_Media_CDDBSlave2_getNTrks (PortableServer_Servant servant,
 
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
 		return -1;
 	}
 
@@ -1081,6 +1252,10 @@ impl_GNOME_Media_CDDBSlave2_getAllTracks (PortableServer_Servant servant,
 	
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
 		return;
 	}
 
@@ -1089,6 +1264,11 @@ impl_GNOME_Media_CDDBSlave2_getAllTracks (PortableServer_Servant servant,
 	(*names_list)->_maximum = entry->ntrks + 1;
 	(*names_list)->_buffer = CORBA_sequence_GNOME_Media_CDDBSlave2_TrackInfo_allocbuf (entry->ntrks);
 
+	if (entry->ntrks == 0) {
+		g_warning ("No tracks\n");
+		return;
+	}
+
 	for (ntrk = 0; ntrk < entry->ntrks; ntrk++) {
 		char *name;
 		GString *result;
@@ -1096,9 +1276,9 @@ impl_GNOME_Media_CDDBSlave2_getAllTracks (PortableServer_Servant servant,
 		name = g_strdup_printf ("TTITLE%d", ntrk);
 		result = g_hash_table_lookup (entry->fields, name);
 		if (result != NULL) {
-			(*names_list)->_buffer[ntrk].name = CORBA_string_dup (result->str ? result->str : "");
+			(*names_list)->_buffer[ntrk].name = CORBA_string_dup (result->str ? result->str : _("Unknown track"));
 		} else {
-			(*names_list)->_buffer[ntrk].name = CORBA_string_dup ("");
+			(*names_list)->_buffer[ntrk].name = CORBA_string_dup (_("Unknown track"));
 		}
 		g_free (name);
 
@@ -1120,12 +1300,59 @@ impl_GNOME_Media_CDDBSlave2_getAllTracks (PortableServer_Servant servant,
 	return;
 }
 
+static void
+set_new_string (CDDBEntry *entry,
+		const char *key,
+		const char *new_contents)
+{
+	GString *dentry;
+
+	dentry = g_hash_table_lookup (entry->fields, key);
+	if (dentry == NULL) {
+		dentry = g_string_new (new_contents);
+		g_hash_table_insert (entry->fields, g_strdup (key), dentry);
+	} else {
+		g_string_printf (dentry, "%s", new_contents);
+	}
+}
+
 static void
 impl_GNOME_Media_CDDBSlave2_setAllTracks (PortableServer_Servant servant,
 					  const CORBA_char *discid,
-					  GNOME_Media_CDDBSlave2_TrackList *list,
+					  const GNOME_Media_CDDBSlave2_TrackList *list,
 					  CORBA_Environment *ev)
 {
+	CDDBEntry *entry;
+	int ntrks, i;
+	
+	entry = g_hash_table_lookup (cddb_cache, discid);
+	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		return;
+	}
+
+	ntrks = list->_length;
+	for (i = 0; i < ntrks; i++) {
+		char *name, *comment, *key;
+		GString *dentry;
+		
+		name = list->_buffer[i].name;
+		comment = list->_buffer[i].comment;
+
+		if (name != NULL && *name != 0) {
+			key = g_strdup_printf ("TTITLE%d", i);
+			set_new_string (entry, key, name);
+			g_free (key);
+		}
+
+		if (comment != NULL && *name != 0) {
+			key = g_strdup_printf ("EXTT%d", i);
+			set_new_string (entry, key, comment);
+			g_free (key);
+		}
+	}
 }
 
 static CORBA_char *
@@ -1138,8 +1365,11 @@ impl_GNOME_Media_CDDBSlave2_getComment (PortableServer_Servant servant,
 	
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
-		/* Must do exceptions...*/
-		return NULL;
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry - %s", discid);
+		return CORBA_string_dup ("");
 	}
 
 	dcomment = g_hash_table_lookup (entry->fields, "EXTD");
@@ -1161,6 +1391,9 @@ impl_GNOME_Media_CDDBSlave2_setComment (PortableServer_Servant servant,
 
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
 		return;
 	}
 
@@ -1183,6 +1416,10 @@ impl_GNOME_Media_CDDBSlave2_getYear (PortableServer_Servant servant,
 
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
 		return -1;
 	}
 	
@@ -1205,6 +1442,9 @@ impl_GNOME_Media_CDDBSlave2_setYear (PortableServer_Servant servant,
 	
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
 		return;
 	}
 
@@ -1228,8 +1468,11 @@ impl_GNOME_Media_CDDBSlave2_getGenre (PortableServer_Servant servant,
 
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
-		/* Exception */
-		return NULL;
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
+		g_warning ("No entry %s", discid);
+		return CORBA_string_dup ("");
 	}
 	
 	dgenre = g_hash_table_lookup (entry->fields, "DGENRE");
@@ -1251,6 +1494,9 @@ impl_GNOME_Media_CDDBSlave2_setGenre (PortableServer_Servant servant,
 
 	entry = g_hash_table_lookup (cddb_cache, discid);
 	if (entry == NULL) {
+		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
+				     ex_GNOME_Media_CDDBSlave2_UnknownDiscID,
+				     NULL);
 		return;
 	}
 
@@ -1276,6 +1522,9 @@ finalize (GObject *object)
 	if (priv == NULL)
 		return;
 
+	/* Remove this slave from our list */
+	cddb_slaves = g_list_remove (cddb_slaves, cddb);
+	
 	g_free (priv->server);
 	g_free (priv->name);
 	g_free (priv->hostname);
@@ -1297,6 +1546,7 @@ cddb_slave_class_init (CDDBSlaveClass *klass)
 
 	parent_class = g_type_class_peek_parent (klass);
 	epv->query = impl_GNOME_Media_CDDBSlave2_query;
+	epv->save = impl_GNOME_Media_CDDBSlave2_save;
 	epv->getArtist = impl_GNOME_Media_CDDBSlave2_getArtist;
 	epv->setArtist = impl_GNOME_Media_CDDBSlave2_setArtist;
 	epv->getDiscTitle = impl_GNOME_Media_CDDBSlave2_getDiscTitle;
@@ -1378,6 +1628,10 @@ cddb_slave_new_full (const char *server,
 	bonobo_object_add_interface (BONOBO_OBJECT (cddb),
 				     BONOBO_OBJECT (priv->event_source));
 
+
+	/* Add this slave to our global list of slaves */
+	cddb_slaves = g_list_prepend (cddb_slaves, cddb);
+	
 	return cddb;
 }
 /**
